==> nono-0.4.4.tar.gz: nono-0.4.4/vm/pedec.cpp <==
//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// PEDEC
//

// IODevice
//    |
//    +-- InterruptDevice (割り込みコントローラの共通部)
//          |
//          +-- M680x0Interrupt (m68k 割り込みコントローラの共通部)
//          |     |
//          |     +-- X68030Interrupt  (X68030 の割り込みコントローラ)
//          |     +-- LunaInterrupt    (LUNA-I の割り込みコントローラ)
//          |
//          |  +-------------+
//          +--| PEDECDevice | (X68030 の Lv1 担当コントローラ)
//          |  +-------------+
//          |
//          +-- SysCtlrDevice  (LUNA-88K の割り込みコントローラ)

#include "pedec.h"
#include "fdc.h"
#include "fdd.h"
#include "spc.h"

// グローバル参照用
PEDECDevice *gPEDEC;

// コンストラクタ
PEDECDevice::PEDECDevice()
	: inherited("PEDEC")
{
}

// デストラクタ
PEDECDevice::~PEDECDevice()
{
	gPEDEC = NULL;
}

// リセット
void
PEDECDevice::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);

	intmap_enabled = IntmapSPC;
	// XXX vector
}

uint64
PEDECDevice::Read(uint32 offset)
{
	switch (offset) {
	 case 0:	// $E9C001 割り込みステータス
		return GetStat();

	 case 1:	// $E9C003 割り込みベクタ
		// 書き込み専用
		return 0xff;

	 default:
		return 0xff;
	}
}

uint64
PEDECDevice::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case 0:	// $E9C001 割り込みマスク
		intmap_enabled = IntmapSPC;
		intmap_enabled |= (data & 0x08) ? IntmapHDD : 0;
		intmap_enabled |= (data & 0x04) ? IntmapFDC : 0;
		intmap_enabled |= (data & 0x02) ? IntmapFDD : 0;
		intmap_enabled |= (data & 0x01) ? IntmapPRT : 0;

		putlog(1, "Set Interrupt Mask HDD=%d FDC=%d FDD=%d PRT=%d",
			((intmap_enabled & IntmapHDD) ? 1 : 0),
			((intmap_enabled & IntmapFDC) ? 1 : 0),
			((intmap_enabled & IntmapFDD) ? 1 : 0),
			((intmap_enabled & IntmapPRT) ? 1 : 0));
		return 0;

	 case 1:	// $E9C003 割り込みベクタ
		vector = data & 0xfc;
		putlog(1, "Set Interrupt Vector $%02x", vector);
		return 0;

	 default:
		return 0;
	}
}

uint64
PEDECDevice::Peek(uint32 offset)
{
	switch (offset) {
	 case 0:
		return GetStat();

	 case 1:
		// 書き込み専用
		return 0xff;

	 default:
		return 0xff;
	}
}

void
PEDECDevice::MonitorUpdate(TextScreen& screen, uint intr_level, int y)
{
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "SPC",	IntmapSPC },
		{ "FDC",	IntmapFDC },
		{ "FDD",	IntmapFDD },
		{ "HDD",	IntmapHDD },
		{ "PRT",	IntmapPRT },
	};

	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		screen.Puts(3, y, TA::OnOff(intmap & map), table[i].name);
		screen.Print(8, y, "%11" PRIu64, counter[clz]);

		if (intr_level >= 1 || (intmap_enabled & map) == 0) {
			screen.Puts(21, y, "Mask");
		}
		y++;
	}
}

// 割り込みステータスレジスタの値を取得。
// (Peek からも呼ばれるので副作用を持たせてはいけない)
uint32
PEDECDevice::GetStat() const
{
	uint32 data = 0;
	data |= (intmap & IntmapFDC) ? 0x80 : 0;
	data |= (intmap & IntmapFDD) ? 0x40 : 0;
	data |= (intmap & IntmapPRT) ? 0x20 : 0;
	data |= (intmap & IntmapHDD) ? 0x10 : 0;
	data |= (intmap_enabled & IntmapHDD) ? 0x08 : 0;
	data |= (intmap_enabled & IntmapFDC) ? 0x04 : 0;
	data |= (intmap_enabled & IntmapFDD) ? 0x02 : 0;
	data |= (intmap_enabled & IntmapPRT) ? 0x01 : 0;
	return data;
}

// 割り込み信号線の状態を変える
void
PEDECDevice::ChangeInterrupt()
{
#if 0 // こんなログ出したいかも
		// ログ表示のためだけ
		if ((intmap_enabled & IntmapFDC)) {
			putlog(1, "FDC からの割り込み要求");
		} else {
			putlog(2, "FDC からの割り込み要求(マスク)");
		}
#endif

	uint32 stat = (intmap & intmap_enabled);
	gInterrupt->ChangeINT(this, stat);
}

// 割り込みアクノリッジ
int
PEDECDevice::InterruptAcknowledge(int lv)
{
	uint32 stat = (intmap & intmap_enabled);

	// SPC と I/O コントローラ担当の4本どっちが優先かは分からないので
	// SPC を優先しておくのでいいだろう。
	if ((stat & IntmapSPC)) {
		// XXX PEDEC のベクタに影響されるか?
		return 0x6c;
	}

	if ((stat & IntmapFDC)) {
		return vector + 0;
	} else if ((stat & IntmapFDD)) {
		return vector + 1;
	} else {
		VMPANIC("Unknown interrupt acknowledge");
	}
}

// デバイスから Intmap 値を引く
uint32
PEDECDevice::GetIntmap(Device *source) const
{
	if (__predict_true(source == gSPC)) {
		return IntmapSPC;
	}
	if (__predict_true(source == gFDC)) {
		return IntmapFDC;
	}
	if (dynamic_cast<FDDDevice*>(source)) {
		return IntmapFDD;
	}
	VMPANIC("Unsupported interrupt device");
}

==> nono-0.6.3.tar.gz: nono-0.6.3/vm/pedec.cpp <==
//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// PEDEC
//

// IODevice
//    |
//    +-- InterruptDevice (割り込みコントローラの共通部)
//          |
//          +-- M680x0Interrupt (m68k 割り込みコントローラの共通部)
//          |     |
//          |     +-- X68030Interrupt  (X68030 の割り込みコントローラ)
//          |     +-- LunaInterrupt    (LUNA-I の割り込みコントローラ)
//          |
//          |  +-------------+
//          +--| PEDECDevice | (X68030 の Lv1 担当コントローラ)
//          |  +-------------+
//          |
//          +-- SysCtlrDevice  (LUNA-88K の割り込みコントローラ)

#include "pedec.h"
#include "fdc.h"
#include "fdd.h"
#include "spc.h"

// コンストラクタ
PEDECDevice::PEDECDevice()
	: inherited(OBJ_PEDEC)
{
}

// デストラクタ
PEDECDevice::~PEDECDevice()
{
}

// 初期化
bool
PEDECDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	fdc = GetFDCDevice();
	interrupt = GetInterruptDevice();
	spc = GetSPCDevice();

	return true;
}

// リセット
void
PEDECDevice::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);

	intmap = IntmapSPC;
	intmap_enabled = IntmapSPC;
	// XXX vector
}

uint64
PEDECDevice::Read(uint32 offset)
{
	switch (offset) {
	 case 0:	// $E9C001 割り込みステータス
		return GetStat();

	 case 1:	// $E9C003 割り込みベクタ
		// 書き込み専用
		return 0xff;

	 default:
		return 0xff;
	}
}

uint64
PEDECDevice::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case 0:	// $E9C001 割り込みマスク
		intmap_enabled = IntmapSPC;
		intmap_enabled |= (data & 0x08) ? IntmapHDD : 0;
		intmap_enabled |= (data & 0x04) ? IntmapFDC : 0;
		intmap_enabled |= (data & 0x02) ? IntmapFDD : 0;
		intmap_enabled |= (data & 0x01) ? IntmapPRT : 0;

		putlog(1, "Set Interrupt Mask HDD=%d FDC=%d FDD=%d PRT=%d",
			((intmap_enabled & IntmapHDD) ? 1 : 0),
			((intmap_enabled & IntmapFDC) ? 1 : 0),
			((intmap_enabled & IntmapFDD) ? 1 : 0),
			((intmap_enabled & IntmapPRT) ? 1 : 0));
		return 0;

	 case 1:	// $E9C003 割り込みベクタ
		vector = data & 0xfc;
		putlog(1, "Set Interrupt Vector $%02x", vector);
		return 0;

	 default:
		return 0;
	}
}

uint64
PEDECDevice::Peek(uint32 offset)
{
	switch (offset) {
	 case 0:
		return GetStat();

	 case 1:
		// 書き込み専用
		return 0xff;

	 default:
		return 0xff;
	}
}

void
PEDECDevice::MonitorUpdate(TextScreen& screen, uint intr_level, int y)
{
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "SPC",	IntmapSPC },
		{ "FDC",	IntmapFDC },
		{ "FDD",	IntmapFDD },
		{ "HDD",	IntmapHDD },
		{ "PRT",	IntmapPRT },
	};

	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		screen.Puts(4, y, TA::OnOff(intmap & map), table[i].name);
		if (intr_level >= 1 || (intmap_enabled & map) == 0) {
			screen.Puts(11, y, "Mask");
		}
		screen.Print(15, y, "%26s", format_number(counter[clz]).c_str());
		y++;
	}
}

// 割り込みステータスレジスタの値を取得。
// (Peek からも呼ばれるので副作用を持たせてはいけない)
uint32
PEDECDevice::GetStat() const
{
	uint32 data = 0;
	data |= (intmap & IntmapFDC) ? 0x80 : 0;
	data |= (intmap & IntmapFDD) ? 0x40 : 0;
	data |= (intmap & IntmapPRT) ? 0x20 : 0;
	data |= (intmap & IntmapHDD) ? 0x10 : 0;
	data |= (intmap_enabled & IntmapHDD) ? 0x08 : 0;
	data |= (intmap_enabled & IntmapFDC) ? 0x04 : 0;
	data |= (intmap_enabled & IntmapFDD) ? 0x02 : 0;
	data |= (intmap_enabled & IntmapPRT) ? 0x01 : 0;
	return data;
}

// 割り込み信号線の状態を変える
void
PEDECDevice::ChangeInterrupt()
{
#if 0 // こんなログ出したいかも
		// ログ表示のためだけ
		if ((intmap_enabled & IntmapFDC)) {
			putlog(1, "FDC からの割り込み要求");
		} else {
			putlog(2, "FDC からの割り込み要求(マスク)");
		}
#endif

	uint32 stat = (intmap & intmap_enabled);
	interrupt->ChangeINT(this, stat);
}

// 割り込みアクノリッジ
int
PEDECDevice::InterruptAcknowledge(int lv)
{
	uint32 stat = (intmap & intmap_enabled);

	// SPC と I/O コントローラ担当の4本どっちが優先かは分からないので
	// SPC を優先しておくのでいいだろう。
	if ((stat & IntmapSPC)) {
		// XXX PEDEC のベクタに影響されるか?
		return 0x6c;
	}

	if ((stat & IntmapFDC)) {
		return vector + 0;
	} else if ((stat & IntmapFDD)) {
		return vector + 1;
	} else {
		VMPANIC("Unknown interrupt acknowledge: intmap=%x enabled=%x",
			intmap, intmap_enabled);
	}
}

// デバイスから Intmap 値を引く
uint32
PEDECDevice::GetIntmap(const Device *source) const
{
	if (__predict_true(source == spc)) {
		return IntmapSPC;
	}
	if (__predict_true(source == fdc)) {
		return IntmapFDC;
	}
	if (dynamic_cast<const FDDDevice*>(source)) {
		return IntmapFDD;
	}
	VMPANIC("Unsupported interrupt device");
}
